"use strict";
/**
 * This Mongo Redis Topic Consumer uses a mongo collection to persistenly represent a message stream.
 * Each document in the collection represents a message within the topic.
 *
 * A redis connection is used to broadcast signals when new messages arrive to unlock waiting consumers.
 * This allows us to avoid a polling mechanism inside the consumer code.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoRedisTopicConsumer = void 0;
const topicConsumer_1 = require("../../topicConsumer");
const mongoClientController_1 = require("./mongoClientController");
const redisClientController_1 = require("./redisClientController");
class MongoRedisTopicConsumer extends topicConsumer_1.TopicConsumer {
    constructor(mongoUrl, databaseName, collectionName, redisUrl) {
        super(collectionName);
        this.client = null;
        this.redis = null;
        this.mongoUrl = mongoUrl;
        this.databaseName = databaseName;
        this.collectionName = collectionName;
        this.isStarting = false;
        this.isStreaming = false;
        this.isPolling = false;
        this.isMoreMessages = true;
        this.isCrashed = false;
        this.redisUrl = redisUrl;
        this.hasCalledNoMoreMessages = false;
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isStarting) {
                throw new Error("Already started");
            }
            this.isStarting = true;
            const mcc = new mongoClientController_1.MongoClientController(this.mongoUrl, this.collectionName + "-consumer");
            yield mcc.start();
            this.client = mcc.getClient();
            // console.log("Mongo client is connected and ready!");
            const rcc = new redisClientController_1.RedisClientController(this.redisUrl, this.collectionName + "-consumer");
            yield rcc.start((connectionState) => __awaiter(this, void 0, void 0, function* () {
                if (connectionState === "ready") {
                    // Resubscribe
                    if (this.redis !== null) {
                        yield this.startListeningForNewMessageSignal();
                        // console.log("Resubscribed to redis channel!");
                    }
                    // Make sure we trigger the poller when redis reconnects
                    this.isMoreMessages = true;
                }
            }));
            this.redis = rcc.getClient();
            // console.log("Redis client is connected and ready !");
            // Handle wake signal
            this.redis.on("message", (channel, message) => {
                this.isMoreMessages = true;
            });
            yield this.startListeningForNewMessageSignal();
            // console.log("Subscribed to redis signal channel!");
        });
    }
    stop() {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            this.isCrashed = true;
            (_a = this.client) === null || _a === void 0 ? void 0 : _a.close();
            (_b = this.redis) === null || _b === void 0 ? void 0 : _b.disconnect();
        });
    }
    getCollection() {
        if (this.client === null) {
            throw new Error("this.client is null");
        }
        return this.client.db(this.databaseName).collection(this.collectionName);
    }
    streamMessagesFrom(callback, lastId = null, queryDoc = {}, noMoreMessages = () => { }) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isStreaming) {
                throw new Error("Already streaming");
            }
            this.isStreaming = true;
            this.callback = callback;
            this.lastId = lastId;
            this.queryDoc = queryDoc;
            // Start a loop here which polls for new messages until there are no more
            const pollingLoop = setInterval(() => __awaiter(this, void 0, void 0, function* () {
                // Stop forever
                if (this.isCrashed) {
                    clearInterval(pollingLoop);
                    return;
                }
                if (!this.isMoreMessages) { // Set to true to unlock the interval
                    if (!this.hasCalledNoMoreMessages) {
                        this.hasCalledNoMoreMessages = true;
                        try {
                            noMoreMessages();
                        }
                        catch (e) {
                            // Ignore but display error
                            console.error(e);
                        }
                    }
                    return;
                }
                if (this.isPolling) {
                    console.warn("Locked, still polling...");
                    return;
                }
                this.isPolling = true;
                try {
                    yield this.pollForNewMessages();
                    this.isPolling = false;
                }
                catch (e) {
                    console.warn(e);
                    this.isPolling = false;
                }
            }), 1000);
        });
    }
    pollForNewMessages() {
        return __awaiter(this, void 0, void 0, function* () {
            // console.log("Polling for new messages");
            const newMessages = yield this.fetchNextMessagesFromMongo(100, this.lastId, this.queryDoc);
            if (newMessages.length === 0) {
                // No more messages
                // console.log("No more messages");
                this.isMoreMessages = false;
            }
            else {
                // Callback these messages
                for (const msg of newMessages) {
                    // Note: If we don't catch this here, it means that polling will fail if the callback throws an error.  
                    // The poll will instantly retry anyway which will cause the consumer to be spammed with the same next message.
                    // To avoid this behaviour, we say that the consumer itself fails and crash the whole instance.
                    try {
                        this.lastId = msg._id;
                        // Convert to generic type of topic message with "id" field instead of "_id"
                        const topicMsg = Object.assign(Object.assign({}, msg), { _id: undefined, id: msg._id });
                        delete topicMsg._id;
                        this.callback(topicMsg);
                    }
                    catch (e) {
                        console.error(e);
                        console.error("The consumer could not handle the message", JSON.stringify(msg, null, 4));
                        console.error("The consumer will not stream any more messages");
                        this.isCrashed = true;
                        return;
                    }
                }
            }
        });
    }
    fetchNextMessagesFromMongo(limit, lastId, queryDoc) {
        return __awaiter(this, void 0, void 0, function* () {
            const col = this.getCollection();
            const thisQuery = {};
            for (const key in queryDoc) {
                thisQuery["payload." + key] = queryDoc[key];
            }
            if (lastId !== null) {
                thisQuery["_id"] = {
                    $gt: lastId
                };
            }
            /**
             * A note about message order.
             * Message order is very important within the realm of event based systems.
             * There is only one order that messages can arrive in, and so there should only be one order that messages are streamed in.
             * It is important that (during writing) we generate a new _id which is incremented every time (auto incremented) so it is unique and retains order.
             * This means floods of messages arriving at once from multiple producers should be handled consistently by the system.
             *
             * Note: Unsorted find order seems to be arbitrary depending on client, so we must explicitly sort by _id.
             **/
            const results = yield col.find(thisQuery).limit(limit).sort({
                _id: 1,
            }).toArray();
            return results;
        });
    }
    startListeningForNewMessageSignal() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.redis === null) {
                throw new Error("this.redis is null");
            }
            yield this.redis.subscribe("TOPIC-" + this.collectionName);
        });
    }
}
exports.MongoRedisTopicConsumer = MongoRedisTopicConsumer;
